# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class ExotelPriority:
    """
    Priorities
    """

    NORMAL = "normal"
    HIGH = "high"


EXOTEL_PRIORITIES = (
    ExotelPriority.NORMAL,
    ExotelPriority.HIGH,
)

EXOTEL_PRIORITY_MAP = {
    # short for 'normal'
    "normal": ExotelPriority.NORMAL,
    # short for 'high'
    "+": ExotelPriority.HIGH,
    "high": ExotelPriority.HIGH,
}


class ExotelEncoding:
    """
    The different encodings supported
    """

    TEXT = "plain"
    UNICODE = "unicode"


class ExotelRegion:
    """
    Regions
    """

    US = "us"

    # India
    IN = "in"


# Exotel APIs
EXOTEL_API_LOOKUP = {
    ExotelRegion.US: "https://api.exotel.com/v1/Accounts/{sid}/Sms/send",
    ExotelRegion.IN: "https://api.in.exotel.com/v1/Accounts/{sid}/Sms/send",
}

# A List of our regions we can use for verification
EXOTEL_REGIONS = (
    ExotelRegion.US,
    ExotelRegion.IN,
)

EXOTEL_SOURCE_RE = re.compile(r"^[A-Z0-9][A-Z0-9.-]{2,15}$", re.I)


def parse_exotel_source(source):
    """Parse an Exotel source as an ExoPhone or approved Sender ID."""
    source = validate_regex(source)
    if not source:
        return None

    result = is_phone_no(source, min_len=9)
    if result:
        return result["full"]

    if EXOTEL_SOURCE_RE.match(source) and (
        not source.isdigit() or len(source) == 6
    ):
        return source

    return None


class NotifyExotel(NotifyBase):
    """
    A wrapper for Exotel Notifications
    """

    # Exotel supports static bulk SMS by accepting an array of To values on
    # the same /Sms/send endpoint.
    default_batch_size = 100

    # The default descriptive name associated with the Notification
    service_name = "Exotel"

    # The services URL
    service_url = "https://exotel.com/"

    # The default protocol
    secure_protocol = "exotel"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/exotel/"

    # The maximum length of the body
    body_maxlen = 2000

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{sid}:{token}@{from_phone}",
        "{schema}://{sid}:{token}@{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "sid": {
                "name": _("Account SID"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "from_phone": {
                "name": _("From Phone No / Sender ID"),
                "type": "string",
                "required": True,
                "regex": (
                    r"^(?:(?=.{3,16}$)(?=.*[A-Z.-])[A-Z0-9][A-Z0-9.-]*"
                    r"|[0-9]{6}|\+?[0-9\s)(+-]{9,})$",
                    "i",
                ),
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
            "sid": {
                "alias_of": "sid",
            },
            "token": {
                "alias_of": "token",
            },
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "map_to": "apikey",
            },
            "key": {
                "alias_of": "apikey",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "unicode": {
                # Unicode characters
                "name": _("Unicode Characters"),
                "type": "bool",
                "default": True,
            },
            "region": {
                "name": _("Region Name"),
                "type": "choice:string",
                "values": EXOTEL_REGIONS,
                "default": ExotelRegion.US,
                "map_to": "region_name",
            },
            "priority": {
                "name": _("Priority"),
                "type": "choice:string",
                "values": EXOTEL_PRIORITIES,
                "default": ExotelPriority.NORMAL,
            },
        },
    )

    def __init__(
        self,
        sid,
        token,
        source,
        targets=None,
        apikey=None,
        batch=None,
        unicode=None,
        priority=None,
        region_name=None,
        **kwargs,
    ):
        """
        Initialize Exotel Object
        """
        super().__init__(**kwargs)

        # Account SID
        self.sid = validate_regex(sid)
        if not self.sid:
            msg = "An invalid Exotel SID ({}) was specified.".format(sid)
            self.logger.warning(msg)
            raise TypeError(msg)

        # API Token (associated with account)
        self.token = validate_regex(token)
        if not self.token:
            msg = "An invalid Exotel API Token ({}) was specified.".format(
                token
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # API Key used as the HTTP Basic Auth username. Older URLs did not
        # carry this separately, so default it to the account SID.
        self.apikey = validate_regex(apikey if apikey else sid)
        if not self.apikey:
            msg = "An invalid Exotel API Key ({}) was specified.".format(
                apikey
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Used for URL generation afterwards only
        self.invalid_targets = []

        # Store our region
        self.region_name = (
            self.template_args["region"]["default"]
            if region_name is None
            else validate_regex(region_name)
        )
        if not self.region_name:
            self.region_name = ""
        else:
            self.region_name = self.region_name.lower()

        if self.region_name not in EXOTEL_REGIONS:
            # Invalid region specified
            msg = "The Exotel region specified ({}) is invalid.".format(
                region_name
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Define whether or not we should set the unicode flag
        self.unicode = (
            self.template_args["unicode"]["default"]
            if unicode is None
            else bool(unicode)
        )

        # Define whether or not we should operate in batch mode
        self.batch = (
            self.template_args["batch"]["default"]
            if batch is None
            else bool(batch)
        )

        #
        # Priority
        #
        if priority is None:
            # Default
            self.priority = self.template_args["priority"]["default"]

        else:
            # Input is a string; attempt to get the lookup from our
            # priority mapping
            self.priority = priority.lower().strip()

            # Allow partial matching against supported priorities:
            # normal, norma, norm, nor, no, n (for normal)
            # high, hig, hi, h (for high)
            result = (
                next(
                    (
                        key
                        for key in EXOTEL_PRIORITY_MAP
                        if key.startswith(self.priority)
                    ),
                    None,
                )
                if self.priority
                else None
            )

            # Now test to see if we got a match
            if not result:
                msg = "An invalid Exotel priority ({}) was specified.".format(
                    priority
                )
                self.logger.warning(msg)
                raise TypeError(msg)

            # store our successfully looked up priority
            self.priority = EXOTEL_PRIORITY_MAP[result]

        # The Source Phone #
        self.source = source

        result = parse_exotel_source(source)
        if not result:
            msg = (
                "The Account (From) Phone # / Sender ID specified "
                "({}) is invalid.".format(source)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our parsed value
        self.source = result

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target, min_len=9)
            if not result:
                self.logger.warning(
                    "Dropped invalid phone # ({}) specified.".format(target),
                )
                self.invalid_targets.append(target)
                continue

            # store valid phone number
            self.targets.append(result["full"])

        if len(self.targets) == 0 and not self.invalid_targets:
            # No targets specified, use our own source.
            self.targets.append(self.source)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """
        Perform Exotel Notification
        """

        if not self.targets:
            # There were no endpoints to notify
            self.logger.warning("There were no Exotel targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # Our authentication
        auth = (self.apikey, self.token)

        # Prepare our payload
        payload = {
            "From": self.source,
            "Body": body,
            "EncodingType": ExotelEncoding.UNICODE
            if self.unicode
            else ExotelEncoding.TEXT,
            "Priority": self.priority,
            # The to gets populated in the loop below
            "To": None,
        }

        # Prepare our targets
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = (
            list(self.targets)
            if batch_size == 1
            else [
                self.targets[index : index + batch_size]
                for index in range(0, len(self.targets), batch_size)
            ]
        )

        # Prepare our notify_url
        notify_url = EXOTEL_API_LOOKUP[self.region_name].format(sid=self.sid)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user(s)
            payload["To"] = target

            p_target = (
                "{} targets".format(len(target))
                if isinstance(target, list)
                else target
            )

            # Some Debug Logging
            self.logger.debug(
                "Exotel POST URL: {} (cert_verify={})".format(
                    notify_url, self.verify_certificate
                )
            )
            self.logger.debug("Exotel Payload: {}".format(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    notify_url,
                    auth=auth,
                    data=payload.copy(),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyExotel.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Exotel notification to {}: "
                        "{}{}error={}.".format(
                            p_target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(
                        "Response Details:\r\n%r", (r.content or b"")[:2000]
                    )

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent Exotel notification to %s.", p_target
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Exotel:%s "
                    "notification.",
                    p_target,
                )
                self.logger.debug("Socket Exception: %s", str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.apikey,
            self.sid,
            self.token,
            self.source,
            self.region_name,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
            "unicode": "yes" if self.unicode else "no",
            "region": self.region_name,
            "priority": self.priority,
        }
        if self.apikey != self.sid:
            params["apikey"] = self.pprint(self.apikey, privacy, safe="")

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{sid}:{token}@{source}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            sid=self.pprint(
                self.sid, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            token=self.pprint(self.token, privacy, safe=""),
            source=NotifyExotel.quote(self.source, safe=""),
            targets="/".join(
                [
                    NotifyExotel.quote(x, safe="")
                    for x in chain(self.targets, self.invalid_targets)
                ]
            ),
            params=NotifyExotel.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if not targets:
            return 1

        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyExotel.split_path(results["fullpath"])

        # The hostname is our source number
        results["source"] = NotifyExotel.unquote(results["host"])

        # Get our account_sid and token from the user/pass config
        results["sid"] = NotifyExotel.unquote(results["user"])
        results["token"] = NotifyExotel.unquote(results["password"])

        # Get region
        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            results["region_name"] = NotifyExotel.unquote(
                results["qsd"]["region"]
            )

        # API Token
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            # Extract the Token from an argument
            results["token"] = NotifyExotel.unquote(results["qsd"]["token"])

        # API SID
        if "sid" in results["qsd"] and len(results["qsd"]["sid"]):
            # Extract the API SID from an argument
            results["sid"] = NotifyExotel.unquote(results["qsd"]["sid"])

        # API Key
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            # Extract the API Key from an argument
            results["apikey"] = NotifyExotel.unquote(results["qsd"]["apikey"])

        elif "key" in results["qsd"] and len(results["qsd"]["key"]):
            # Extract the API Key from an argument
            results["apikey"] = NotifyExotel.unquote(results["qsd"]["key"])

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyExotel.unquote(results["qsd"]["from"])
        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyExotel.unquote(results["qsd"]["source"])

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyExotel.parse_phone_no(
                results["qsd"]["to"]
            )

        # Get priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyExotel.unquote(
                results["qsd"]["priority"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyExotel.template_args["batch"]["default"]
            )
        )

        # Unicode Characters
        results["unicode"] = parse_bool(
            results["qsd"].get(
                "unicode", NotifyExotel.template_args["unicode"]["default"]
            )
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# You must generate a "Long-Lived Access Token". This can be done from your
# Home Assistant Profile page.
from itertools import chain
from json import dumps
import math
import re
from uuid import uuid4

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import (
    is_domain_service_target,
    parse_bool,
    parse_domain_service_targets,
    validate_regex,
)
from .base import NotifyBase

# This regex matches exactly 8 hex digits,
# a dot, then exactly 64 hex digits. it can also be a JWT
# token in which case it will be 180 characters+
RE_IS_LONG_LIVED_TOKEN = re.compile(
    r"^([0-9a-f]{8}\.[0-9a-f]{64}|[a-z0-9_-]+\.[a-z0-9_-]+\.[a-z0-9_-]+)$",
    re.I,
)

# Define our supported device notification formats:
# - service
#     - default domain is always 'notify' if one isn't detected
# - service:target
# - service:target1,target2,target3
# - domain.service
# - domain.service:target
# - domain.service:target1,target2,target3
# -   - targets can be comma/space separated if more hten one
# - service:target1,target2,target3

# Define a persistent entry (used for handling message delivery
PERSISTENT_ENTRY = (None, None, [])


class NotifyHomeAssistant(NotifyBase):
    """
    A wrapper for Home Assistant Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = "HomeAssistant"

    # The services URL
    service_url = "https://www.home-assistant.io/"

    # Insecure Protocol Access
    protocol = "hassio"

    # Secure Protocol
    secure_protocol = "hassios"

    # Default to Home Assistant Default Insecure port of 8123 instead of 80
    default_insecure_port = 8123

    # The maximum amount of services that can be notified in a single batch
    default_batch_size = 10

    # The default ha notification domain if one isn't detected
    default_domain = "notify"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/homeassistant/"

    # Define object templates
    templates = (
        "{schema}://{host}/{accesstoken}",
        "{schema}://{host}/{accesstoken}/{targets}",
        "{schema}://{host}:{port}/{accesstoken}",
        "{schema}://{host}:{port}/{accesstoken}/{targets}",
        "{schema}://{user}@{host}/{accesstoken}",
        "{schema}://{user}@{host}/{accesstoken}/{targets}",
        "{schema}://{user}@{host}:{port}/{accesstoken}",
        "{schema}://{user}@{host}:{port}/{accesstoken}/{targets}",
        "{schema}://{user}:{password}@{host}/{accesstoken}",
        "{schema}://{user}:{password}@{host}/{accesstoken}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{accesstoken}",
        "{schema}://{user}:{password}@{host}:{port}/{accesstoken}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "accesstoken": {
                "name": _("Long-Lived Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "nid": {
                # Optional Unique Notification ID
                "name": _("Notification ID"),
                "type": "string",
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "to": {
                "alias_of": "targets",
            },
            "token": {
                # Shorthand alias for accesstoken in query string
                "alias_of": "accesstoken",
            },
            "prefix": {
                "name": _("Path Prefix"),
                "type": "string",
            },
        },
    )

    def __init__(
        self,
        accesstoken,
        nid=None,
        targets=None,
        batch=None,
        prefix=None,
        **kwargs,
    ):
        """Initialize Home Assistant Object."""
        super().__init__(**kwargs)

        self.fullpath = kwargs.get("fullpath", "")
        self.prefix = prefix if prefix else ""

        if not (self.secure or self.port):
            # Use default insecure port
            self.port = self.default_insecure_port

        # Long-Lived Access token (generated from User Profile)
        self.accesstoken = validate_regex(accesstoken)
        if not self.accesstoken:
            msg = (
                "An invalid Home Assistant Long-Lived Access Token "
                f"({accesstoken}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # An Optional Notification Identifier
        self.nid = None
        if nid:
            self.nid = validate_regex(nid, *self.template_args["nid"]["regex"])
            if not self.nid:
                msg = (
                    "An invalid Home Assistant Notification Identifier "
                    f"({nid}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        # Prepare Batch Mode Flag
        self.batch = (
            self.template_args["batch"]["default"] if batch is None else batch
        )

        # Store our targets
        self.targets = []

        # Track our invalid targets
        self._invalid_targets = []

        if targets:
            for target in parse_domain_service_targets(targets):
                result = is_domain_service_target(
                    target, domain=self.default_domain
                )
                if result:
                    self.targets.append(
                        (
                            result["domain"],
                            result["service"],
                            result["targets"],
                        )
                    )
                    continue

                self.logger.warning(
                    "Dropped invalid [domain.]service[:target] entry "
                    "({}) specified.".format(target),
                )
                self._invalid_targets.append(target)
        else:
            self.targets = [PERSISTENT_ENTRY]

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Sends Message."""

        # Base payload; notification_id is only added for persistent
        # notification calls — other service domains reject it.
        payload = {
            "title": title,
            "message": body,
        }

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.accesstoken}",
        }

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        # Determine if we're doing it the old way (using persistent notices)
        # or the new (supporting device targets)
        has_targets = bool(
            not self.targets or self.targets[0] is not PERSISTENT_ENTRY
        )

        if has_targets and not self.targets:
            self.logger.warning(
                "No valid Home Assistant service targets to notify."
            )
            return False

        base_url = url + (
            self.fullpath.rstrip("/")
            + "/api/services/persistent_notification/create"
        )

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        for target in self.targets:
            # Use a unique ID so we don't over-write the last message we
            # posted. Otherwise use the notification id specified
            if has_targets:
                # Base target details
                domain = target[0]
                service = target[1]

                # Prepare our URL
                base_url = (
                    url
                    + self.prefix.rstrip("/")
                    + f"/api/services/{domain}/{service}"
                )

                # Possibly prepare batches
                if target[2]:
                    _payload = payload.copy()
                    for index in range(0, len(target[2]), batch_size):
                        _payload["targets"] = target[2][
                            index : index + batch_size
                        ]
                        if not self._ha_post(
                            base_url, _payload, headers, auth
                        ):
                            return False

                    continue

            if not self._ha_post(
                base_url,
                payload,
                headers,
                auth,
                persistent=not has_targets,
            ):
                return False

        return True

    def _ha_post(self, url, payload, headers, auth=None, persistent=False):
        """
        Wrapper to single upstream server post
        """
        # notification_id is only meaningful for persistent_notification;
        # other HA service domains reject it with a 400.
        if persistent:
            payload["notification_id"] = self.nid if self.nid else str(uuid4())

        self.logger.debug(
            "Home Assistant POST URL: {} (cert_verify={!r})".format(
                url, self.verify_certificate
            )
        )
        self.logger.debug("Home Assistant Payload: {}".format(str(payload)))

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyHomeAssistant.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Home Assistant notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(
                    "Response Details:\r\n%r", (r.content or b"")[:2000]
                )

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Home Assistant notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Home Assistant "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.fullpath.rstrip("/"),
            self.accesstoken,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
        }

        if self.prefix not in ("", "/"):
            params["prefix"] = (
                "/"
                if not self.prefix
                else "/{}/".format(self.prefix.strip("/"))
            )

        if self.nid:
            params["nid"] = self.nid

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyHomeAssistant.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyHomeAssistant.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else self.default_insecure_port

        url = (
            "{schema}://{auth}{hostname}{port}{fullpath}"
            "{accesstoken}/{targets}?{params}"
        )

        # Determine if we're doing it the old way (using persistent notices)
        # or the new (supporting device targets)
        has_targets = bool(
            not self.targets or self.targets[0] is not PERSISTENT_ENTRY
        )

        return url.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if not self.port or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=(
                "/"
                if not self.fullpath
                else "/{}/".format(
                    NotifyHomeAssistant.quote(
                        self.fullpath.strip("/"), safe="/"
                    )
                )
            ),
            accesstoken=self.pprint(self.accesstoken, privacy, safe=""),
            targets=""
            if not has_targets
            else "/".join(
                chain(
                    [
                        NotifyHomeAssistant.quote(
                            "{}.{}{}".format(
                                x[0],
                                x[1],
                                "" if not x[2] else ":" + ",".join(x[2]),
                            ),
                            safe="",
                        )
                        for x in self.targets
                    ],
                    [
                        NotifyHomeAssistant.quote(x, safe="")
                        for x in self._invalid_targets
                    ],
                )
            ),
            params=NotifyHomeAssistant.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        #
        # Factor batch into calculation
        #

        # Determine if we're doing it the old way (using persistent notices)
        # or the new (supporting device targets)
        has_targets = bool(
            not self.targets or self.targets[0] is not PERSISTENT_ENTRY
        )

        if not has_targets:
            return 1

        # Handle targets
        batch_size = 1 if not self.batch else self.default_batch_size
        return sum(
            math.ceil(len(identities) / batch_size) if identities else 1
            for _, _, identities in self.targets
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Initialize targets
        results["targets"] = []

        # Get our Long-Lived Access Token — check query string first
        # (?accesstoken= or the shorthand ?token=)
        if "accesstoken" in results["qsd"] and results["qsd"]["accesstoken"]:
            results["accesstoken"] = NotifyHomeAssistant.unquote(
                results["qsd"]["accesstoken"]
            )
            # Remaining path elements become targets
            results["targets"] = NotifyHomeAssistant.split_path(
                results["fullpath"]
            )
            results["fullpath"] = ""

        elif "token" in results["qsd"] and results["qsd"]["token"]:
            results["accesstoken"] = NotifyHomeAssistant.unquote(
                results["qsd"]["token"]
            )
            # Remaining path elements become targets
            results["targets"] = NotifyHomeAssistant.split_path(
                results["fullpath"]
            )
            results["fullpath"] = ""

        else:
            # Scan path forward; the first element matching the token
            # regex is the access token.  Everything before it is the
            # prefix path; everything after it (reversed for consistent
            # call ordering) is treated as service targets.
            tokens = NotifyHomeAssistant.split_path(results["fullpath"])
            token_idx = None
            for idx, t in enumerate(tokens):
                if RE_IS_LONG_LIVED_TOKEN.match(t):
                    token_idx = idx
                    results["accesstoken"] = t
                    break

            if token_idx is not None:
                # Elements after the token are service targets (reversed
                # so the last URL segment is called first).
                post_targets = list(reversed(tokens[token_idx + 1 :]))
                # Elements before the token are also treated as service
                # targets (reversed to maintain consistent ordering).
                pre_targets = list(reversed(tokens[:token_idx]))
                results["targets"] = post_targets + pre_targets
                results["fullpath"] = ""

            elif tokens:
                # No regex match — treat the last path element as the
                # access token (plain token like "accesstoken" in tests).
                results["accesstoken"] = tokens[-1]
                results["fullpath"] = ""

            else:
                results["accesstoken"] = None

        # Support ?to= for additional targets
        if "to" in results["qsd"] and results["qsd"]["to"]:
            results["targets"] += NotifyHomeAssistant.split_path(
                NotifyHomeAssistant.unquote(results["qsd"]["to"])
            )

        # Support ?prefix= path-prefix override
        if "prefix" in results["qsd"] and results["qsd"]["prefix"]:
            results["prefix"] = NotifyHomeAssistant.unquote(
                results["qsd"]["prefix"]
            )

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch",
                NotifyHomeAssistant.template_args["batch"]["default"],
            )
        )

        # Allow the specification of a unique notification_id so that
        # it will always replace the last one sent.
        if "nid" in results["qsd"] and len(results["qsd"]["nid"]):
            results["nid"] = NotifyHomeAssistant.unquote(results["qsd"]["nid"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""A minimal IRC client for notification delivery."""

from __future__ import annotations

from collections import deque
import logging
import random
import string
import time
from typing import Callable, Optional, Union

from ...logger import logger
from ...utils.socket import AppriseSocketError, SocketTransport
from .protocol import (
    IRCAuthMode,
    is_ping,
    normalise_channel,
    parse_irc_line,
    ping_payload,
)
from .state import IRCActionKind, IRCContext, IRCStateMachine

NickGenerator = Callable[[str, int, int], str]


class IRCClient:
    """Socket-driven IRC Client."""

    # IRC Default Ports
    default_insecure_port = 6667
    default_secure_port = 6697

    # When we're blocking and awaiting the server, we increase the
    # frequency to poll for an update.
    pump_interval = 0.75

    # EFNet only allows nicknames to be 9 characters, default to that.
    nickname_max_length = 9

    # How many times we retry on 432/433 before failing.
    nickname_collision_max = 3

    def __init__(
        self,
        host: str,
        nickname: str,
        fullname: str,
        port: Optional[int] = None,
        secure: bool = False,
        verify: bool = True,
        timeout: Optional[float] = None,
        password: Optional[str] = None,
        auth_mode: str = IRCAuthMode.SERVER,
        nick_generator: Optional[NickGenerator] = None,
        nick_length: Optional[int] = None,
    ) -> None:

        # Detect port if not set
        port = (
            port
            if port is not None
            else (
                self.default_secure_port
                if secure
                else self.default_insecure_port
            )
        )

        self.transport = SocketTransport(
            host,
            port,
            secure=secure,
            verify=verify,
            timeout=timeout,
        )

        self._nick_generator = nick_generator
        self._nick_length = (
            int(nick_length) if nick_length else int(self.nickname_max_length)
        )
        self._nick_collision = 0

        self.auth_mode = auth_mode

        ctx = IRCContext(
            desired_nick=nickname,
            accepted_nick=nickname,
            fullname=fullname,
            password=password,
        )
        self.sm = IRCStateMachine(ctx)
        self._out_queue = deque()  # type: Deque[bytes]
        self._inbuf = bytearray()

    @property
    def nickname(self) -> str:
        """Returns the accepted nickname."""
        return self.sm.ctx.accepted_nick

    def connect(self) -> None:
        self.transport.connect()

    def close(self) -> None:
        self.transport.close()

    def _queue(self, line: str) -> None:
        """
        Queues message for outbound delivery to the IRC Server
        """
        payload = (line + "\r\n").encode("utf-8", errors="replace")
        self._out_queue.append(payload)

    def _write(self, line: Union[str, bytes], deadline: float) -> None:
        """Write content directly to IRC."""
        remaining = max(0.0, deadline - time.monotonic())
        if remaining <= 0.0:
            raise TimeoutError("timeout while writing IRC commands")

        payload = (
            (line + "\r\n").encode("utf-8", errors="replace")
            if isinstance(line, str)
            else line
        )
        self.transport.write(payload, flush=True, timeout=remaining)
        if logger.isEnabledFor(logging.TRACE):
            logger.trace(
                "IRC write: %s",
                payload.rstrip(b"\r")
                .decode("utf-8", errors="replace")
                .rstrip(),
            )

    def _flush(self, deadline: float) -> None:
        """Flush all queued information to the IRC server."""
        while self._out_queue:
            self._write(self._out_queue[0], deadline=deadline)
            self._out_queue.popleft()

    def _read(self, deadline: float) -> Optional[str]:
        """
        Read incoming content from IRC Server
        """
        while True:
            if b"\n" in self._inbuf:
                line, _, rest = self._inbuf.partition(b"\n")
                self._inbuf = bytearray(rest)
                response = line.rstrip(b"\r").decode("utf-8", errors="replace")
                logger.trace("IRC read: %s", response)
                return response

            remaining = max(0.0, deadline - time.monotonic())
            if remaining <= 0.0:
                logger.trace("IRC read timeout - deadline=%.2fs", deadline)
                return None

            chunk = self.transport.read(4096, blocking=True, timeout=remaining)
            if not chunk:
                return None
            self._inbuf.extend(chunk)

    def _nickname_collision_handler(self, prefix: str) -> str:
        if not self._nick_generator:
            raise AppriseSocketError("Nickname collision and no generator")

        if self._nick_collision >= int(self.nickname_collision_max):
            raise AppriseSocketError("Nickname is already in use")

        self._nick_collision += 1
        self.sm.ctx.desired_nick = self._nick_generator(
            prefix,
            self._nick_length,
            self._nick_collision,
        )
        return self.sm.ctx.desired_nick

    def _tick(self, deadline: float) -> float:
        remaining = max(0.0, deadline - time.monotonic())
        if remaining <= 0.0:
            return deadline
        return time.monotonic() + min(self.pump_interval, remaining)

    def _handshake(self, deadline: float, prefix: str) -> None:
        while True:
            line = self._read(deadline=deadline)
            if not line:
                # We've completed
                return

            msg = parse_irc_line(line)  # type: IRCMessage

            if is_ping(msg):
                self._write(f"PONG :{ping_payload(msg)}", deadline=deadline)
                continue

            if msg.numeric in (432, 433):
                new_nick = self._nickname_collision_handler(prefix)
                # Send immediately, do not queue.
                self._write("NICK {}".format(new_nick), deadline=deadline)
                continue

            for act in self.sm.on_message(msg):
                if act.kind == IRCActionKind.FAIL and act.reason:
                    raise AppriseSocketError(act.reason)
                if act.kind == IRCActionKind.SEND and act.line:
                    self._write(act.line, deadline=deadline)

    def register(self, timeout: float, prefix: str) -> None:
        """Register with the IRC server, and optionally NickServ identify.

        - SERVER mode: sends PASS during registration (if password provided)
        - NICKSERV mode: does not send PASS, performs NickServ IDENTIFY after
          registration completes
        - NONE mode: no authentication is performed
        """
        tl_start = time.time()
        deadline = time.monotonic() + float(timeout)

        # PASS during registration is only used for SERVER and ZNC.
        if self.auth_mode not in (IRCAuthMode.SERVER, IRCAuthMode.ZNC):
            self.sm.ctx.password = None

        logger.trace("IRC registration started")
        for act in self.sm.start_registration():
            if act.kind == IRCActionKind.SEND and act.line:
                self._queue(act.line)

        while time.monotonic() < deadline and not self.sm.ctx.registered:
            self._flush(deadline)
            self._handshake(self._tick(deadline), prefix=prefix)

        if not self.sm.ctx.registered:
            logger.trace(
                "IRC registration timeout - %.6fs elapsed",
                time.time() - tl_start,
            )
            raise TimeoutError("IRC registration timeout")

        logger.trace(
            "IRC registration completed in %.6fs",
            time.time() - tl_start,
        )

        # NickServ identify is only performed after we are registered, and only
        # when explicitly requested via auth_mode.
        if self.auth_mode == IRCAuthMode.NICKSERV:
            self.identify(timeout=timeout)

    def check_connection(self, timeout: float) -> bool:
        """Verify we can talk to the server by completing a PING/PONG."""
        deadline = time.monotonic() + float(timeout)
        token = "apprise"

        # Send a ping and wait until we observe a PONG carrying our token.
        self._write(f"PING :{token}", deadline=deadline)

        while time.monotonic() < deadline:
            line = self._read(deadline=deadline)
            if not line:
                continue

            msg = parse_irc_line(line)
            if msg.command.upper() == "PONG":
                # Some IRC servers/bouncers do not echo our token back
                # reliably. Observing any PONG after issuing our PING is
                # sufficient.
                return True

        return False

    def join(
        self,
        channel: str,
        timeout: float,
        prefix: str,
        key: Optional[str] = None,
    ) -> None:

        chan = normalise_channel(channel)
        if chan in self.sm.ctx.joined:
            # Nothing to do, we are already there.
            return
        deadline = time.monotonic() + float(timeout)

        for act in self.sm.request_join(chan, key=key):
            if act.kind == IRCActionKind.SEND and act.line:
                self._queue(act.line)

        while time.monotonic() < deadline and chan not in self.sm.ctx.joined:
            self._flush(deadline)
            self._handshake(self._tick(deadline), prefix=prefix)

        if chan not in self.sm.ctx.joined:
            logger.debug("IRC join confirmation not observed for %s", chan)

    def privmsg(self, target: str, message: str, timeout: float) -> None:
        """Handle the sending of private messages."""
        deadline = time.monotonic() + float(timeout)
        self._queue(f"PRIVMSG {target} :{message}")
        self._flush(deadline)
        self._handshake(self._tick(deadline), prefix="")

    def identify(self, timeout: float) -> None:
        """Identify with NickServ after registration."""
        if not self.sm.ctx.password:
            return

        if self.auth_mode != IRCAuthMode.NICKSERV:
            return

        deadline = time.monotonic() + float(timeout)
        self._queue(
            "PRIVMSG NickServ :IDENTIFY {}".format(self.sm.ctx.password)
        )
        self._flush(deadline)
        self._handshake(
            self._tick(deadline),
            prefix="",
        )

    def quit(self, message: str, timeout: float) -> None:
        deadline = time.monotonic() + float(timeout)
        for act in self.sm.request_quit(message):
            if act.kind == IRCActionKind.SEND and act.line:
                self._queue(act.line)
        self._flush(deadline)

    @staticmethod
    def nick_generation(
        prefix: str, length: Optional[int] = None, collision: int = 0
    ) -> str:
        """Generate a nickname suitable for retry after collision."""
        if length is None:
            # Default Assignment
            length = IRCClient.nickname_max_length

        base = "{}".format(prefix)[: length - 3].strip().lower()
        charset = string.ascii_lowercase + string.digits + "_"
        suffix = "".join(random.choice(charset) for _ in range(max(1, length)))
        nick = "{}{}".format(base, suffix)
        if collision:
            nick = "{}{}".format(nick[: max(0, length - 1)], collision % 10)
        return nick[:length]



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Sources:
# - https://blink1.thingm.com/
# - https://github.com/todbot/blink1-python
# - https://github.com/todbot/blink1/blob/main/docs/blink1-hid-commands.md

import time

from ..common import NotifyType
from ..locale import gettext_lazy as _
from .base import NotifyBase

# Default our global support flag
NOTIFY_BLINK1_ENABLED = False

try:
    import hid as blink1_hid

    NOTIFY_BLINK1_ENABLED = True

except ImportError:
    # No problem -- hidapi is an optional dependency.  Users who wish to
    # use this plugin must install it:  pip install hidapi
    pass

#
# blink(1) USB HID constants
#

# USB vendor / product identifiers for all blink(1) revisions
BLINK1_VENDOR_ID = 0x27B8
BLINK1_PRODUCT_ID = 0x01ED

# HID report size (report ID byte + 8 payload bytes)
BLINK1_REPORT_ID = 0x01
BLINK1_REPORT_SIZE = 9

# Command byte: 'c' == fade to RGB with a time argument
BLINK1_CMD_FADE = ord("c")


class Blink1LED:
    """
    Defines the LED to focus on
    """

    ALL = 0
    FIRST = 1
    SECOND = 2


BLINK1_LED_CHOICES = {
    Blink1LED.ALL: "all",
    Blink1LED.FIRST: "1",
    Blink1LED.SECOND: "2",
}

# Maps inbound strings to Blink1LED constants; unknown values fall back to ALL
BLINK1_LED_MAP = {
    "0": Blink1LED.ALL,
    "all": Blink1LED.ALL,
    "a": Blink1LED.ALL,
    "1": Blink1LED.FIRST,
    "2": Blink1LED.SECOND,
}

# How long (ms) to hold the LED colour after the fade completes
BLINK1_DEFAULT_DURATION_MS = 5000
BLINK1_MIN_DURATION_MS = 0
BLINK1_MAX_DURATION_MS = 300000  # 5 minutes

# Fade transition time (ms); 0 = instant
BLINK1_DEFAULT_FADE_MS = 0
BLINK1_MIN_FADE_MS = 0
BLINK1_MAX_FADE_MS = 10000  # 10 seconds


def _blink1_fade_buf(red, green, blue, fade_ms, ledn):
    """Build the 9-byte HID feature-report buffer for a fade-to-RGB command.

    The blink(1) wire format (all values unsigned 8-bit unless noted):
      [0] REPORT_ID (0x01)
      [1] command 'c' (0x63)
      [2] red
      [3] green
      [4] blue
      [5] fade_time high byte  (fade_time = fade_ms // 10, 16-bit big-endian)
      [6] fade_time low byte
      [7] ledn (0=all, 1=LED1, 2=LED2)
      [8] 0x00 (padding)
    """
    fade_time = int(fade_ms) // 10
    th = (fade_time >> 8) & 0xFF
    tl = fade_time & 0xFF
    return [
        BLINK1_REPORT_ID,
        BLINK1_CMD_FADE,
        int(red),
        int(green),
        int(blue),
        th,
        tl,
        int(ledn),
        0x00,
    ]


class NotifyBlink1(NotifyBase):
    """A wrapper for blink(1) USB LED notifications.

    Colors are driven by Apprise's notification-type color map
    (info=blue, success=green, warning=yellow, failure=red).
    No external blink1 library is required; the USB HID wire protocol
    is implemented directly via the hidapi package.
    """

    # Set our global enabled flag
    enabled = NOTIFY_BLINK1_ENABLED

    requirements = {
        "packages_required": "hidapi",
    }

    # The default descriptive name associated with the notification
    service_name = _("blink(1)")

    # The services URL
    service_url = "https://blink1.thingm.com/"

    # The default protocol
    protocol = "blink1"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/blink1/"

    # blink(1) is a local USB device; a title field has no meaning here.
    title_maxlen = 0

    # No throttling needed for a local USB device
    request_rate_per_sec = 0

    # URL templates
    templates = (
        "{schema}://",
        "{schema}://{serial}/",
    )

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "serial": {
                "name": _("Serial Number"),
                "type": "string",
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "duration": {
                "name": _("Duration (ms)"),
                "type": "int",
                "min": BLINK1_MIN_DURATION_MS,
                "max": BLINK1_MAX_DURATION_MS,
                "default": BLINK1_DEFAULT_DURATION_MS,
            },
            "fade": {
                "name": _("Fade Time (ms)"),
                "type": "int",
                "min": BLINK1_MIN_FADE_MS,
                "max": BLINK1_MAX_FADE_MS,
                "default": BLINK1_DEFAULT_FADE_MS,
            },
            "ledn": {
                "name": _("LED Number"),
                "type": "choice:int",
                "values": BLINK1_LED_CHOICES,
                "default": Blink1LED.ALL,
            },
        },
    )

    def __init__(
        self,
        serial=None,
        duration=None,
        fade=None,
        ledn=None,
        **kwargs,
    ):
        """Initialize Blink1 Object."""

        super().__init__(**kwargs)

        # Device serial number; None means "first available device".
        # An underscore is accepted as a URL placeholder meaning "any".
        serial = serial.strip() if isinstance(serial, str) else None
        self.serial = serial if serial and serial != "_" else None

        # Duration (ms) to hold the LED colour before turning it off
        try:
            self.duration = int(
                BLINK1_DEFAULT_DURATION_MS if duration is None else duration
            )
            if not (
                BLINK1_MIN_DURATION_MS
                <= self.duration
                <= BLINK1_MAX_DURATION_MS
            ):
                raise ValueError("out of range")

        except (TypeError, ValueError):
            msg = (
                f"blink(1) duration ({duration}) must be between"
                f" {BLINK1_MIN_DURATION_MS} and"
                f" {BLINK1_MAX_DURATION_MS} ms."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # Fade transition time (ms)
        try:
            self.fade = int(BLINK1_DEFAULT_FADE_MS if fade is None else fade)
            if not (BLINK1_MIN_FADE_MS <= self.fade <= BLINK1_MAX_FADE_MS):
                raise ValueError("out of range")

        except (TypeError, ValueError):
            msg = (
                f"blink(1) fade ({fade}) must be between"
                f" {BLINK1_MIN_FADE_MS} and {BLINK1_MAX_FADE_MS} ms."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # LED selector; unrecognised values silently fall back to ALL
        self.ledn = (
            BLINK1_LED_MAP.get(str(ledn).lower(), Blink1LED.ALL)
            if ledn is not None
            else Blink1LED.ALL
        )

    def _open_device(self):
        """Open and return a hidapi device handle for the blink(1).

        Returns None and logs a warning when the device cannot be found.
        """
        try:
            dev = blink1_hid.device()
            dev.open(
                BLINK1_VENDOR_ID,
                BLINK1_PRODUCT_ID,
                self.serial,
            )
            return dev

        except OSError:
            msg = "Failed to open blink(1) device"
            if self.serial:
                msg += f" (serial={self.serial})"
            msg += "."
            self.logger.warning(msg)
            return None

    def _send_fade(self, dev, red, green, blue, fade_ms):
        """Send a single fade-to-RGB HID feature report.

        Returns True on success, False if the report could not be sent.
        """
        buf = _blink1_fade_buf(red, green, blue, fade_ms, self.ledn)
        try:
            rc = dev.send_feature_report(buf)

        except OSError as e:
            self.logger.warning("blink(1) HID write failed: %s", e)
            return False

        if rc != BLINK1_REPORT_SIZE:
            self.logger.warning(
                "blink(1) HID write returned %d (expected %d).",
                rc,
                BLINK1_REPORT_SIZE,
            )
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all fields that uniquely identify this connection."""
        return (
            self.protocol,
            self.serial,
            self.ledn,
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform blink(1) Notification."""

        # Resolve the RGB triple for this notification type
        r, g, b = self.color(notify_type=notify_type, color_type=tuple)

        dev = self._open_device()
        if dev is None:
            return False

        try:
            # Always throttle before any device I/O
            self.throttle()

            if not self._send_fade(dev, r, g, b, self.fade):
                return False

            # Hold for fade + duration, then switch off
            time.sleep((self.fade + self.duration) / 1000.0)

            # Turn off: instant fade to black
            if not self._send_fade(dev, 0, 0, 0, 0):
                return False

        finally:
            dev.close()

        self.logger.info("Sent blink(1) notification.")
        return True

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = {
            "duration": str(self.duration),
            "fade": str(self.fade),
            "ledn": str(self.ledn),
        }
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        serial = self.serial if self.serial else "_"
        return f"{self.protocol}://{serial}/?{NotifyBlink1.urlencode(params)}"

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments to re-instantiate."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            return results

        # The hostname, when present, is the device serial number.
        # An underscore or empty value means "first available device".
        host = results.get("host", "")
        if host and host != "_":
            results["serial"] = NotifyBlink1.unquote(host)

        if results["qsd"].get("duration"):
            results["duration"] = NotifyBlink1.unquote(
                results["qsd"]["duration"]
            )

        if results["qsd"].get("fade"):
            results["fade"] = NotifyBlink1.unquote(results["qsd"]["fade"])

        if results["qsd"].get("ledn"):
            results["ledn"] = NotifyBlink1.unquote(results["qsd"]["ledn"])

        return results

    @staticmethod
    def runtime_deps():
        """Return optional runtime dependency package names."""
        return ("hid",)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Utilities that make payloads safe to *print* in debug and trace logs.

This module is intentionally scoped to log presentation only.
"""

from __future__ import annotations

from dataclasses import dataclass
from hashlib import sha256
from typing import Any, Optional

# Keys that commonly contain large binary-like values.
#
# This list is used as a *hint* to enable a more aggressive summarization mode
# for values underneath matching keys. The structure is still walked, only the
# leaf string/bytes values are summarized more aggressively.
#
# Keep this list conservative. Very generic tokens may cause false positives.
_BLOB_KEYWORDS = (
    "base64",
    "attachment",
    "attachments",
    "base64_attachments",
    "contentbytes",
    "blob",
    "file",
    "data",
    "image",
    "media",
    "document",
)


@dataclass(frozen=True)
class SanitizeOptions:
    """Options controlling payload sanitization for debug logging.

    The defaults are deliberately conservative and tuned for logging, not for
    data processing. When in doubt, prefer smaller values to keep logging
    cheap.

    Attributes:
        max_depth: Maximum recursion depth before truncation markers appear.
        max_items: Global upper bound on visited items, across the whole walk.
        max_str_len: Strings longer than this are summarized with a preview.
        preview: Number of characters to show at the start and end of
                 summaries.
        hash_sample_size: Maximum bytes hashed when generating a sha256
                          preview.
        aggressive_blob_keys: If True, summarize values under blob-like keys
            even when they are not huge, because these values are often encoded
            attachments (for example, base64).
    """

    # How many recursive lists/sets/tuples/dicts to delve into before
    # aborting
    max_depth: int = 10

    # The max amount of fields to process before we just abort (too many)
    max_items: int = 100

    # Strings longer than this are summarized
    max_str_len: int = 512

    # Preview size for summarized strings
    preview: int = 32

    # Bound hashing work
    # Strings longer than this (usually large attachments) include
    # sha256 hash value of it in response
    hash_sample_size: int = 8192

    # If True, summarize values under blob-like keys even if they are smaller
    # than the defined max_str_len.
    aggressive_blob_keys: bool = True


def sanitize_payload(
    value: Any, *, options: Optional[SanitizeOptions] = None
) -> Any:
    """
    This function is intended for DEBUG and TRACE logging only.

    can add i/o to generate the printed copy, but the output is much
    better then just printing what could be a massive payload (with
    attachments).

    The ideal setup for this function is when you need to print what
    could be a very large object such as in the send() of a Apprise
    service, you would structure it like this:

        # check for at least the DEBUG level... you can also set
        # logging.TRACE if you wanted as well:
        if self.logger.isEnabledFor(logging.DEBUG):

            # Then safely wrap the output using this function:
            self.logger.debug(
                "Service Payload: %s", sanitize_payload(payload))
    """
    opts = options or SanitizeOptions()

    # Track already-seen objects to prevent infinite loops on recursive graphs.
    seen_ids: set[int] = set()

    # Global counter that enforces opts.max_items across the entire traversal.
    items_seen = 0

    def _hash_bytes(b: bytes) -> str:
        """Return a short sha256 prefix for bytes, bounded by
        hash_sample_size."""
        if len(b) > opts.hash_sample_size:
            b = b[: opts.hash_sample_size]
        return sha256(b).hexdigest()[:12]

    def _summarize_str(s: str) -> str:
        """Summarize strings longer than opts.max_str_len with a preview."""
        nonlocal items_seen
        items_seen += 1

        length = len(s)
        if length <= opts.max_str_len:
            return s

        head = s[: opts.preview]
        tail = s[-opts.preview :] if length >= opts.preview else s
        return f"<string len={length} head={head!r} tail={tail!r}>"

    def _summarize_bytes(b: bytes) -> str:
        """Summarize bytes with length and a short bounded sha256 prefix."""
        nonlocal items_seen
        items_seen += 1
        return f"<bytes len={len(b)} sha256={_hash_bytes(b)}>"

    def _is_blob_key(k: str) -> bool:
        """Return True if a key name indicates blob-like content.

        This keeps behaviour predictable while avoiding heavy scanning of
        values. Add items to _BLOB_KEYWORDS (not case sensitive) if required
        to optimize speed of parsing.
        """
        lk = k.lower()
        return lk in _BLOB_KEYWORDS

    def _summarize_key(k: Any) -> Any:
        """Summarize keys where needed, preserving readability in logs."""
        if isinstance(k, str):
            # Keys are usually short. Only summarize if they are unexpectedly
            # large to avoid log pollution.
            return _summarize_str(k)
        if isinstance(k, bytes):
            return _summarize_bytes(k)
        return k

    def _walk(obj: Any, depth: int, *, blob_mode: bool = False) -> Any:
        """Recursively walk payload structures and summarize leaf values."""
        nonlocal items_seen

        # Global safety limits first, so we can exit cheaply.
        if items_seen >= opts.max_items:
            return "<truncated: global item limit reached>"
        if depth > opts.max_depth:
            return "<truncated: max depth reached>"

        # Pass-through primitives.
        if obj is None or isinstance(obj, (bool, int, float)):
            items_seen += 1
            return obj

        # Strings: optionally apply blob-mode summaries.
        if isinstance(obj, str):
            if blob_mode and opts.aggressive_blob_keys:
                # Always summarize blob fields, even if not huge
                length = len(obj)
                head = obj[: opts.preview]
                tail = obj[-opts.preview :] if length >= opts.preview else obj
                items_seen += 1
                return (
                    f"<string len={length} blob head={head!r} tail={tail!r}>"
                )
            return _summarize_str(obj)

        # Bytes: always summarize.
        if isinstance(obj, bytes):
            return _summarize_bytes(obj)

        # Prevent recursion loops on self-referential objects.
        obj_id = id(obj)
        if obj_id in seen_ids:
            return "<recursive>"
        seen_ids.add(obj_id)

        # Dict: walk values. Keys may be summarized for readability.
        if isinstance(obj, dict):
            out: dict[Any, Any] = {}
            for k, v in obj.items():
                if items_seen >= opts.max_items:
                    out["<truncated>"] = "..."
                    break

                sk = _summarize_key(k)

                # Enable blob mode for suspicious keys. We still walk the
                # structure, but leaf values are summarized aggressively.
                child_blob_mode = blob_mode
                if isinstance(k, str) and _is_blob_key(k):
                    child_blob_mode = True

                out[sk] = _walk(v, depth + 1, blob_mode=child_blob_mode)
            return out

        # Sequences: walk each entry. Sets are returned as lists for
        #            readability.
        if isinstance(obj, (list, tuple, set, frozenset)):
            out_list: list[Any] = []
            for entry in obj:
                if items_seen >= opts.max_items:
                    out_list.append("<truncated: limit reached>")
                    break
                out_list.append(_walk(entry, depth + 1, blob_mode=blob_mode))

            if isinstance(obj, tuple):
                return tuple(out_list)
            return out_list

        # Unknown objects: fall back to repr().
        items_seen += 1
        return repr(obj)

    # Recursively walk all elements of the object passed
    return _walk(value, 0)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class Singleton(type):
    """Our Singleton MetaClass."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Instantiate our singleton meta entry."""
        if cls not in cls._instances:
            # we have not every built an instance before.  Build one now.
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a httpSMS account
# You will need credits (new accounts start with a few)
#     https://httpsms.com
import json

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from .base import NotifyBase


class NotifyHttpSMS(NotifyBase):
    """A wrapper for HttpSMS Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "httpSMS"

    # The services URL
    service_url = "https://httpsms.com"

    # All notification requests are secure
    secure_protocol = "httpsms"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/httpsms/"

    # HttpSMS uses the http protocol with JSON requests
    notify_url = "https://api.httpsms.com/v1/messages/send"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{apikey}@{from_phone}",
        "{schema}://{apikey}@{from_phone}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
                "required": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "key": {
                "alias_of": "apikey",
            },
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
        },
    )

    def __init__(self, apikey=None, source=None, targets=None, **kwargs):
        """Initialize HttpSMS Object."""
        super().__init__(**kwargs)

        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = (
                f"The Account (From) Phone # specified ({source}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Tidy source
        self.source = result["full"]

        # Parse our targets
        self.targets = []

        has_error = False
        for target in parse_phone_no(targets):
            # Parse each phone number we found
            result = is_phone_no(target)
            if result:
                self.targets.append(result["full"])
                continue

            has_error = True
            self.logger.warning(
                f"Dropped invalid phone # ({target}) specified.",
            )

        if not targets and not has_error:
            # Default the SMS Message to ourselves
            self.targets.append(self.source)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform HttpSMS Notification."""

        if not self.targets:
            # We have nothing to notify
            self.logger.warning("There are no HttpSMS targets to notify")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "x-api-key": self.apikey,
            "Content-Type": "application/json",
        }

        # Prepare our payload
        payload = {
            # The To gets populated in the loop below
            "from": "+" + self.source,
            "to": None,
            "content": body,
        }

        # Prepare our targets
        targets = list(self.targets)
        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload["to"] = "+" + target

            # Some Debug Logging
            self.logger.debug(
                "HttpSMS POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"HttpSMS Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=json.dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code
                    )

                    # set up our status code to use
                    status_code = r.status_code

                    self.logger.warning(
                        "Failed to send HttpSMS notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            status_code,
                        )
                    )

                    self.logger.debug(
                        "Response Details:\r\n%r", (r.content or b"")[:2000]
                    )

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent HttpSMS notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending HttpSMS: to %s ",
                    target,
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.source, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # A nice way of cleaning up the URL length a bit
        targets = (
            []
            if len(self.targets) == 1 and self.targets[0] == self.source
            else self.targets
        )

        return "{schema}://{apikey}@{source}/{targets}?{params}".format(
            schema=self.secure_protocol,
            source=self.source,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join(
                [NotifyHttpSMS.quote(f"{x}", safe="+") for x in targets]
            ),
            params=NotifyHttpSMS.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""

        return len(self.targets) if self.targets else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our API Key
        results["apikey"] = NotifyHttpSMS.unquote(results["user"])

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyHttpSMS.unquote(results["qsd"]["from"])

            # hostname will also be a target in this case
            results["targets"] = [
                *NotifyHttpSMS.parse_phone_no(results["host"]),
                *NotifyHttpSMS.split_path(results["fullpath"]),
            ]

        else:
            # store our source
            results["source"] = NotifyHttpSMS.unquote(results["host"])

            # store targets
            results["targets"] = NotifyHttpSMS.split_path(results["fullpath"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyHttpSMS.parse_phone_no(
                results["qsd"]["to"]
            )

        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            results["apikey"] = NotifyHttpSMS.unquote(results["qsd"]["key"])

        return results
